(* $Id: godi_ptool.ml 711 2007-07-30 22:37:57Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2005 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_versions
open Godi_plist
open Godi_exec
open Printf

let sum root files =
  (* Files relative to root *)
  List.fold_left 
    (fun acc file ->
       Int64.add acc (Godi_file.file_size (Filename.concat root file)))
    0L
    files
;;

let has_symlink_component root name_s =

  let rec check root path =
    match path with
      | [] -> 
	  false
      | [lastname] ->
	  false   (* We do not check here whether the whole path exists... *)
      | dir :: path' ->
	  let root' =
	    Filename.concat root dir in
	  if Godi_file.file_type ~is_symlink:true root' then
	    true
	  else
	    check root' path'
  in

  match name_s with
    | Godi_file.Absolute path ->
	let root = Godi_file.root in
	check root path
    | Godi_file.Relative path ->
	check root path
;;



class inst_plist_dispersion plist : dispersion =
  let root = Godi_plist.extract_prefix plist in
  let files = ref [] in
  let core_plist = ref [] in
  let () =
    List.iter
      (function
	 | `File(path, chksum) ->
	     files := path :: !files;
	     core_plist := (`File(path, chksum)) :: !core_plist
	 | #core_plist_element as e ->
	     core_plist := e :: !core_plist
	 | _ ->
	     ()
      )
      plist;
    files := List.rev !files;
    core_plist := List.rev !core_plist
  in
  let size = sum root !files in
object
  method root = Some root
  method files = !files
  method size = size
  method plist = !core_plist
end


let inst_plist_dispersion plist = 
  new inst_plist_dispersion plist ;;



let library_re_config =
  (* Note: These are extended REs, not Str-REs! *)
  [ "ELF", 
    ( fun name ->
	sprintf "(^|/)lib%s\\.(so(\\.[0-9.]+)?|a|la)$" name );

    "a.out",
    ( fun name ->
	sprintf "(^|/)lib%s\\.(so(\\.[0-9.]+)?|a|la)$" name );

    "PE",
    ( fun name ->
	sprintf "(^|/)(lib%s\\.(a|la)|cyg%s\\.dll|%s\\.dll|%s\\.lib)$"
	  name name name name);

    "MachO",
    ( fun name ->
	sprintf "(^|/)lib%s\\.(([0-9]+\\.)+dylib|a|la)$" name );
    
    "SOM",
    ( fun name ->
	sprintf "(^|/)lib%s\\.([0-9]+|sl|a|la)$" name );

    "NA",
    ( fun name ->
	sprintf "(^|/)lib%s\\.(a|la)$" name );
  ] ;;

let stubdll_re_config =
  (* Note: These are extended REs, not Str-REs! *)
  [ "ELF", 
    ( fun name -> 
	sprintf "(^|/)dll%s\\.so(\\.owner)?$" name );

    "a.out",
    ( fun name -> 
	sprintf "(^|/)dll%s\\.so(\\.owner)?$" name );

    "PE",
    ( fun name ->
	sprintf "(^|/)dll%s\\.dll(\\.owner)?$" name );

    "MachO",
    ( fun name ->
	sprintf "(^|/)dll%s\\.so(\\.owner)?$" name );

    "SOM",
    ( fun name ->
	sprintf "(^|/)dll%s\\.sl(\\.owner)?$" name );

    "NA",
    ( fun name ->
	sprintf "(^|/)dll%s\\.so(\\.owner)?$" name );
  ] ;;


class user_plist_dispersion ?(root = Godi_query.localbase()) plist
      : dispersion =

  let shlib_type = Godi_query.shlib_type() in

  let library_re_expr =
    try List.assoc shlib_type library_re_config
    with
      | Not_found -> List.assoc "NA" library_re_config in

  let stubdll_re_expr =
    try List.assoc shlib_type stubdll_re_config
    with
      | Not_found -> List.assoc "NA" stubdll_re_config in

  let library_re name = Godi_match.extended_regexp (library_re_expr name) in
  let stubdll_re name = Godi_match.extended_regexp (stubdll_re_expr name) in

  let core_plist =
    Godi_plist.expand_plist
      ~root
      ~ocamlopt_feature:(Godi_query.have_ocamlopt())
      ~ocamloptp_feature:(Godi_query.have_ocamloptp())
      ~ocamldll_feature:(Godi_query.have_ocamldll())
      ~library_re
      ~stubdll_re
      ~cygwin_fixups:false (* TODO *)
      plist in
  let files =
    List.flatten
      (List.map
	 (function
	    | `File(name,_) ->
		(* Absolute file names are not allowed! *)
		(* TODO: Move these checks to expand_plist *)
		let name_s = Godi_file.split_filename name in
		( match name_s with
		    | Godi_file.Absolute _ ->
			failwith ("Cannot process absolute filename: " ^ name)
		    | Godi_file.Relative l ->
			if List.mem ".." l then
			  failwith ("Not allowed to use .. in PLIST paths: " ^
				      name);
			(* but "." is no problem! *)
		);
		if has_symlink_component root name_s then
		  failwith("Not allowed to cross symlinks in PLIST paths: " ^ 
			     name);
		[name]
	    | _ -> [])
	 core_plist) in
  let size = sum root files in
object
  method root = Some root
  method files = files
  method size = size
  method plist = core_plist
end


let user_plist_dispersion ?root plist =
  new user_plist_dispersion ?root plist ;;


class stage_dispersion dir : dispersion =
  let files = Godi_file.read_dir_hierarchy dir in
  let size = sum dir files in
  let plist =
    List.map
      (fun file ->
	 let p = Filename.concat dir file in
	 if Godi_file.file_type ~is_symlink:true p then
	   `File(file, None)
	 else
	   let sum = "MD5:" ^ Digest.to_hex (Digest.file p) in
	   `File(file, Some sum)
      )
      files
  in
object
  method root = Some dir
  method files = files
  method size = size
  method plist = plist
end


let stage_dispersion dir =
  new stage_dispersion dir ;;


let rec n_times n s =
  if n = 0 then
    []
  else
    s :: n_times (n-1) s


let fixup_symlinks con d =
  let disp_root =
    match d # root with
      | Some d -> Godi_file.split_filename d
      | None -> failwith "Missing root directory for dispersion" in
  let _localbase =
    Godi_file.split_filename (Godi_query.localbase()) in
  List.iter
    (fun file ->
       let file_s = Godi_file.split_filename file in
       let file_n =  (* number of components other than "." *)
	 List.length
	   (List.filter 
	      (fun comp -> comp <> ".")
	      ( match file_s with
		  | Godi_file.Absolute _ -> assert false
		  | Godi_file.Relative p -> p)
	   ) in
       assert(file_n >= 1);

       let abs_file = 
	 Filename.concat 
	   (Godi_file.combine_filename disp_root) file in

       try
	 let link = 
	   Godi_file.split_filename 
	     (Godi_file.read_symlink abs_file) in
		
	 let new_link_opt =
	   ( try
	       let disp_root_suffix = 
		 Godi_file.strip_prefix ~prefix:disp_root link in
	       let up_path =
		 n_times (file_n-1) ".." in  (* TODO: portability *)
	       let new_link =
		 up_path @ disp_root_suffix in
	       Some new_link
	     with
	       | Godi_file.Not_a_prefix ->
		   None
	   ) in
	 ( match new_link_opt with
	     | None -> ()
	     | Some new_link ->
		 printmsg con ("Fixing symlink " ^ abs_file ^ "\n");
		 Godi_file.remove abs_file;
		 let new_link_s =
		   String.concat "/" new_link in    (* TODO: portability *)
		 Godi_file.create_symlink new_link_s abs_file
	 )
       with
	 | Godi_file.Not_a_symlink -> ()
    )
    d#files
;;


class create_dispersed_package
        ?protocol
        ~name ~version ~pkgrevision ~comment ~description ~godi_maintainer
        ~dependencies ~build_dependencies ~conflicts ~install_message
        ~build_version ~build_info ~preserve ~requirements_script
        ~install_script ~deinstall_script (d : dispersion) con
         : dispersed_package =
  let tmp = Godi_file.allocate_tmpdir() in
  let tmpdir = Godi_file.tmpdir_path tmp in

  (* First compute the closure of [dependencies], i.e. all requirements.
   * Second, take the exact version numbers instead of the input conditions
   *)
  let installed_pkgs = Godi_query.get_installed_packages con in
  let required_pkgs = 
    List.map 
      (fun ((pkg_name, _) as dep) -> 
	 try
	   let pkg = 
	     Godi_query.get_this_installed_package con pkg_name in
	   if not (Godi_query.matches_dep pkg dep) then
	     user_error ("Package does not fulfil dependency: " ^ 
			   Godi_versions.print_dependency dep);
	   pkg
	 with
	   | Not_found ->
	       user_error ("Required package not found: " ^ 
			     Godi_versions.print_dependency dep)
      )
      dependencies in
  let required_pkgs' =
    Godi_plan.sort_packages
      (required_pkgs @
	 Godi_plan.get_required_packages
	 ~active:installed_pkgs 
	 ~selected:required_pkgs
	 ()) in
  let dependencies' =
    List.map
      (fun pkg ->
	 let n = pkg_name pkg in
	 let v = pkg_version pkg in
	 Godi_versions.parse_dependency (n ^ "==" ^ v)
      )
      required_pkgs' in
  let req_size =
    List.fold_left
      (fun acc pkg -> Int64.add acc ((installed_pkg pkg) # size))
      0L
      required_pkgs' in
  let build_dependencies' =
    List.map
      (fun ((pkg_name, _) as dep) -> 
	 try
	   let pkg = 
	     Godi_query.get_this_installed_package con pkg_name in
	   if Godi_query.matches_dep pkg dep then
	     Godi_versions.parse_dependency (pkg_name ^ "==" ^
					       (pkg_version pkg))
	   else
	     dep
	 with
	   | Not_found -> dep
      )
      build_dependencies in

  (* Check [conflicts]: *)
  let () =
    List.iter
      (fun ((n, _) as dep) ->
	 try
	   let pkg = 
	     Godi_query.get_this_installed_package con n in
	   if Godi_query.matches_dep pkg dep then
	     user_error ("Conflicting package found: " ^ 
			   Godi_versions.print_dependency dep)
	 with
	   | Not_found -> ()
      )
      conflicts in

  (* Now determine "+CONTENTS": *)
  let inst_plist =
    [ `Name (Godi_versions.concat_name_version name version pkgrevision);
      `Cwd (Godi_query.localbase());
    ] @
      (List.map (fun dep -> `Pkgdep dep) dependencies') @
      (List.map (fun dep -> `Blddep dep) build_dependencies') @
      (List.map (fun dep -> `Pkgcfl dep) conflicts) @
      (match install_message with
	 | "" -> []
	 | _ -> [ `Display "+DISPLAY" ]
      ) @ ( d # plist :> inst_plist_element list ) in

  (* Write the files into [tmpdir]: *)
  let write = Godi_file.write_file ~mode:`Text in
  let () =
    Godi_plist.print_inst_plist (Filename.concat tmpdir "+CONTENTS") inst_plist;
    List.iter 
      (fun (file_name, file_contents) -> 
	 write (Filename.concat tmpdir file_name) file_contents)
      [ "+DESC", description;
	"+COMMENT", comment;
	"+MAINTAINER", godi_maintainer
      ];
    List.iter
      (fun (file_name, file_contents_opt) ->
	 match file_contents_opt with
	   | None -> ()
	   | Some c -> write (Filename.concat tmpdir file_name) c
      )
      [ "+REQUIRE", requirements_script;
	"+INSTALL", install_script;
	"+DEINSTALL", deinstall_script;
	"+DISPLAY", 
	  (if install_message <> "" then Some install_message else None);
	"+BUILD_VERSION", 
	  (if build_version <> "" then Some build_version else None);
	"+BUILD_INFO", 
	  (if build_info <> "" then Some build_info else None)
      ];
    if preserve then
      write (Filename.concat tmpdir "+PRESERVE") 
	"This package must not be deleted\n";
  in
object(self : 'self)
  method name = name
  method version = version
  method pkgrevision = pkgrevision
  method comment = comment
  method description = description
  method godi_maintainer = godi_maintainer
  method dependencies = dependencies
  method build_dependencies = build_dependencies
  method conflicts = conflicts
  method user_status = `Keep
  method set_user_status s =
    if s <> `Keep then failwith "Bad user_status"
  method var_id = new_var_id()
  method variant (v:var_id) : 'self =
    failwith "class dispersed_package: No such variant"
  method dispersion = d
  method tmp_directory = tmp
  method install_message = install_message
  method build_version = build_version
  method build_info = build_info
  method preserve = preserve
  method inst_plist = inst_plist
  method requirements_script = requirements_script
  method install_script = install_script
  method deinstall_script = deinstall_script

  method size = d#size
  method size_all = Int64.add req_size d#size
end

let create_dispersed_package
        ~name ~version ~pkgrevision ~comment ~description ~godi_maintainer
        ~dependencies ~build_dependencies ~conflicts ~install_message
        ~build_version ~build_info ~preserve ~requirements_script
        ~install_script ~deinstall_script (d : dispersion) con =
  `Dispersed_pkg
    (new create_dispersed_package
        ~name ~version ~pkgrevision ~comment ~description ~godi_maintainer
        ~dependencies ~build_dependencies ~conflicts ~install_message
        ~build_version ~build_info ~preserve ~requirements_script
        ~install_script ~deinstall_script (d : dispersion) con
    )
;;


let dispersed_of_installed_package con pkg =
  let ipkg = installed_pkg pkg in
  let d = inst_plist_dispersion ipkg # inst_plist in
  create_dispersed_package
    ~name:(pkg_name pkg)
    ~version:(pkg_version pkg)
    ~pkgrevision:(pkg_pkgrevision pkg)
    ~comment:(pkg_comment pkg)
    ~description:(pkg_description pkg)
    ~godi_maintainer:(pkg_godi_maintainer pkg)
    ~dependencies:(pkg_dependencies pkg)
    ~build_dependencies:(pkg_build_dependencies pkg)
    ~conflicts:(pkg_conflicts pkg)
    ~install_message:ipkg#install_message
    ~build_version:ipkg#build_version
    ~build_info:ipkg#build_info
    ~preserve:ipkg#preserve
    ~requirements_script:ipkg#requirements_script
    ~install_script:ipkg#install_script
    ~deinstall_script:ipkg#deinstall_script
    d con
;;


let binary_of_dispersed_package
      con out_file pkg =
  let dpkg = dispersed_pkg pkg in
  match dpkg # dispersion # root with
    | None ->
	failwith "Godi_ptool.binary_of_dispersed_package"
    | Some root ->
	(* First write a file with the contents, every line is a file name.
         * The files of the tmp directory are written as absolute paths
         * and renamed later. The package files are written as relative
         * paths.
         *)
	let tmpdir = Godi_file.tmpdir_path (dpkg # tmp_directory) in
	let meta_files = Godi_file.read_dir tmpdir in
	let meta_files' = List.filter (fun file -> file.[0] = '+') meta_files in

	exec_to_user_error
	  ~pref_user_error:true
	  (Godi_pax.create_archive
	     ~in_files:[ tmpdir, meta_files';
			 root, dpkg # dispersion # files
		       ]
	     ~out_file
	  )
	  con;

	Godi_query.get_binary_package out_file
;;
